/**
 * Copyright 2018 Google Inc. All Rights Reserved.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *     http://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

const TAG_SET = 1;
const PROPS_SET = 2;
const PROPS_ASSIGN = 3;
const CHILD_RECURSE = 4;
const CHILD_APPEND = 0;

const MODE_SLASH = 0;
const MODE_TEXT = 1;
const MODE_WHITESPACE = 2;
const MODE_TAGNAME = 3;
const MODE_ATTRIBUTE = 4;

const evaluate = (h, current, fields, args) => {
	for (let i = 1; i < current.length; i++) {
		const field = current[i++];
		const value = typeof field === 'number' ? fields[field] : field;

		if (current[i] === TAG_SET) {
			args[0] = value;
		}
		else if (current[i] === PROPS_SET) {
			(args[1] = args[1] || {})[current[++i]] = value;
		}
		else if (current[i] === PROPS_ASSIGN) {
			args[1] = Object.assign(args[1] || {}, value);
		}
		else if (current[i]) {
			// code === CHILD_RECURSE
			args.push(h.apply(null, evaluate(h, value, fields, ['', null])));
		}
		else {
			// code === CHILD_APPEND
			args.push(value);
		}
	}

	return args;
};

const build = function(statics) {

	let mode = MODE_TEXT;
	let buffer = '';
	let quote = '';
	let current = [0];
	let char, propName;

	const commit = field => {
		if (mode === MODE_TEXT && (field || (buffer = buffer.replace(/^\s*\n\s*|\s*\n\s*$/g,'')))) {
			{
				current.push(field || buffer, CHILD_APPEND);
			}
		}
		else if (mode === MODE_TAGNAME && (field || buffer)) {
			{
				current.push(field || buffer, TAG_SET);
			}
			mode = MODE_WHITESPACE;
		}
		else if (mode === MODE_WHITESPACE && buffer === '...' && field) {
			{
				current.push(field, PROPS_ASSIGN);
			}
		}
		else if (mode === MODE_WHITESPACE && buffer && !field) {
			{
				current.push(true, PROPS_SET, buffer);
			}
		}
		else if (mode === MODE_ATTRIBUTE && propName) {
			{
				current.push(field || buffer, PROPS_SET, propName);
			}
			propName = '';
		}
		buffer = '';
	};

	for (let i=0; i<statics.length; i++) {
		if (i) {
			if (mode === MODE_TEXT) {
				commit();
			}
			commit(i);
		}

		for (let j=0; j<statics[i].length; j++) {
			char = statics[i][j];

			if (mode === MODE_TEXT) {
				if (char === '<') {
					// commit buffer
					commit();
					{
						current = [current];
					}
					mode = MODE_TAGNAME;
				}
				else {
					buffer += char;
				}
			}
			else if (quote) {
				if (char === quote) {
					quote = '';
				}
				else {
					buffer += char;
				}
			}
			else if (char === '"' || char === "'") {
				quote = char;
			}
			else if (char === '>') {
				commit();
				mode = MODE_TEXT;
			}
			else if (!mode) ;
			else if (char === '=') {
				mode = MODE_ATTRIBUTE;
				propName = buffer;
				buffer = '';
			}
			else if (char === '/') {
				commit();
				if (mode === MODE_TAGNAME) {
					current = current[0];
				}
				mode = current;
				{
					(current = current[0]).push(mode, CHILD_RECURSE);
				}
				mode = MODE_SLASH;
			}
			else if (char === ' ' || char === '\t' || char === '\n' || char === '\r') {
				// <a disabled>
				commit();
				mode = MODE_WHITESPACE;
			}
			else {
				buffer += char;
			}
		}
	}
	commit();
	return current;
};

const getCacheMap = (statics) => {
	let tpl = CACHE.get(statics);
	if (!tpl) {
		CACHE.set(statics, tpl = build(statics));
	}
	return tpl;
};

const getCacheKeyed = (statics) => {
	let key = '';
	for (let i = 0; i < statics.length; i++) {
		key += statics[i].length + '-' + statics[i];
	}
	return CACHE[key] || (CACHE[key] = build(statics));
};

const USE_MAP =  typeof Map === 'function';
const CACHE = USE_MAP ? new Map() : {};
const getCache = USE_MAP ? getCacheMap : getCacheKeyed;

const cached = function(statics) {
	const res = evaluate(this, getCache(statics), arguments, []);
	return res.length > 1 ? this(res) : res[0];
};

var index =  cached;

/**
 * @param {Babel} babel
 * @param {object} options
 * @param {string} [options.pragma=h]  JSX/hyperscript pragma.
 * @param {string} [options.tag=html]  The tagged template "tag" function name to process.
 * @param {boolean} [options.monomorphic=false]  Output monomorphic inline objects instead of using String literals.
 * @param {boolean} [options.useBuiltIns=false]  Use the native Object.assign instead of trying to polyfill it.
 * @param {boolean} [options.variableArity=true] If `false`, always passes exactly 3 arguments to the pragma function.
 */
function htmBabelPlugin({ types: t }, options = {}) {
	const pragma = options.pragma===false ? false : (options.pragma || 'h|hs');
	const useBuiltIns = options.useBuiltIns;
	const inlineVNodes = options.monomorphic || pragma===false;
	const symbol = Symbol();
  let currentPragma;

	function dottedIdentifier(keypath) {
		const path = keypath.split('.');
		let out;
		for (let i=0; i<path.length; i++) {
			const ident = propertyName(path[i]);
			out = i===0 ? ident : t.memberExpression(out, ident);
		}
		return out;
	}

	function patternStringToRegExp(str) {
		const parts = str.split('/').slice(1);
		const end = parts.pop() || '';
		return new RegExp(parts.join('/'), end);
	}

	function propertyName(key) {
		if (key.match(/(^\d|[^a-z0-9_$])/i)) return t.stringLiteral(key);
		return t.identifier(key);
	}

	function stringValue(str) {
		if (options.monomorphic) {
			return t.objectExpression([
				t.objectProperty(propertyName('type'), t.numericLiteral(3)),
				t.objectProperty(propertyName('tag'), t.nullLiteral()),
				t.objectProperty(propertyName('props'), t.nullLiteral()),
				t.objectProperty(propertyName('children'), t.nullLiteral()),
				t.objectProperty(propertyName('text'), t.stringLiteral(str))
			]);
		}
		return t.stringLiteral(str);
	}

	function createVNode(tag, props, children) {
		// Never pass children=[[]].
		if (children.elements.length === 1 && t.isArrayExpression(children.elements[0]) && children.elements[0].elements.length === 0) {
			children = children.elements[0];
		}

		if (inlineVNodes) {
			return t.objectExpression([
				options.monomorphic && t.objectProperty(propertyName('type'), t.numericLiteral(1)),
				t.objectProperty(propertyName('tag'), tag),
				t.objectProperty(propertyName('props'), props),
				t.objectProperty(propertyName('children'), children),
				options.monomorphic && t.objectProperty(propertyName('text'), t.nullLiteral())
			].filter(Boolean));
		}

		// Passing `{variableArity:false}` always produces `h(tag, props, children)` - where `children` is always an Array.
		// Otherwise, the default is `h(tag, props, ...children)`.
		if (options.variableArity !== false) {
			children = children.elements;
		}

		return t.callExpression(currentPragma, [tag, props].concat(children));
	}

	function flatten(props, result = []) {
		const { [symbol]: head, ...tail } = props;
		if (head) head.forEach(obj => {
			flatten(obj, result);
		});
		if (Object.keys(tail).length > 0) {
			result.push(tail);
		}
		return result;
	}

	function spreadNode(args, state) {
		if (args.length > 0 && t.isNode(args[0])) {
			args.unshift({});
		}

		// 'Object.assign(x)', can be collapsed to 'x'.
		if (args.length === 1) {
			return propsNode(args[0]);
		}
		// 'Object.assign({}, x)', can be collapsed to 'x'.
		if (args.length === 2 && !t.isNode(args[0]) && Object.keys(args[0]).length === 0) {
			return propsNode(args[1]);
		}
		const helper = useBuiltIns ? dottedIdentifier('Object.assign') : state.addHelper('extends');
		return t.callExpression(helper, args.map(propsNode));
	}

	function propsNode(props) {
		return t.isNode(props) ? props : t.objectExpression(
			Object.keys(props).map(key => {
				let value = props[key];
				if (typeof value==='string') {
					value = t.stringLiteral(value);
				}
				else if (typeof value==='boolean') {
					value = t.booleanLiteral(value);
				}
				return t.objectProperty(propertyName(key), value);
			})
		);
	}

	function transform(node, state) {
		if (node === undefined) return t.identifier('undefined');
		if (node == null) return t.nullLiteral();
    if (t.isNode(node)) return node;

		const { tag, props, children } = node;
		function childMapper(child) {
			if (typeof child==='string') {
				return stringValue(child);
			}
			return t.isNode(child) ? child : transform(child, state);
		}

		const newTag = typeof tag === 'string' ? t.stringLiteral(tag) : tag;
		const newProps = props ? spreadNode(flatten(props), state) : t.nullLiteral();
		const newChildren = t.arrayExpression((children || []).map(childMapper));
		return createVNode(newTag, newProps, newChildren);
	}

	function h(tag, props, ...children) {
    if (Array.isArray(tag)) {
      return tag;
    }
		return { tag, props, children };
	}

	const html = index.bind(h);

	function treeify(statics, expr) {
		const assign = Object.assign;
		try {
			Object.assign = function(...objs) {	return { [symbol]: objs }; };
			return html(statics, ...expr);
		}
		finally {
			Object.assign = assign;
		}
	}

	// The tagged template tag function name we're looking for.
	// This is static because it's generally assigned via htm.bind(h),
	// which could be imported from elsewhere, making tracking impossible.
	const htmlName = options.tag || '/html|svg/';
	return {
		name: 'htm',
		visitor: {
			TaggedTemplateExpression(path, state) {
				const tag = path.node.tag.name;

        let match = tag === htmlName;
        let matchName = htmlName;
        if (htmlName[0]==='/') {
          match = tag.match(patternStringToRegExp(htmlName));
          matchName = match[0];
        }

				if (match) {
          const matchIndex = htmlName.replace(/\//g, '').split('|').indexOf(matchName);
          currentPragma = dottedIdentifier(pragma.split('|')[matchIndex]);

					const statics = path.node.quasi.quasis.map(e => e.value.raw);
					const expr = path.node.quasi.expressions;

					const tree = treeify(statics, expr);
					const node = !Array.isArray(tree)
						? transform(tree, state)
						: t.callExpression(currentPragma, [t.arrayExpression(tree.map(root => transform(root, state)))]);
					path.replaceWith(node);
				}
			}
		}
	};
}

module.exports = htmBabelPlugin;
